/*
	saori.h
    Multi Platform SAORI Framework
*/

#pragma once
#define _WINDOWS

#include <string>
#include <vector>
#include <map>
#include <locale>
#include <sstream>

#include <boost/scoped_ptr.hpp>
#include <boost/function.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/type_traits.hpp>

// vbgtH[
#ifdef _WINDOWS

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#define SAORIAPI extern "C" __declspec(dllexport)
#define SAORICALL __cdecl

// Windows ł͋I Unicode g
#define _SAORI_UNICODE

// AP[^̒ۉ
#define SAORI_ALLOC(size) GlobalAlloc(GMEM_FIXED, size)
#define SAORI_FREE(handle) GlobalFree(handle)

#else

#define SAORIAPI extern "C"
#define SAORICALL

// AP[^̒ۉ
#define SAORI_ALLOC(size) malloc(size)
#define SAORI_FREE(handle) free(handle)

#endif /* _WINDOWS */

// S Unicode ŏ邩
#ifdef _SAORI_UNICODE

#define _T(text) L##text

typedef wchar_t char_t;

#ifdef _MSC_VER

// VC  Unicode n̊֐
#define wtoi _wtoi

#endif /* _MSC_VER */

#else

#define _T(text) text

typedef char char_t;

#endif /* _SAORI_UNICODE */

// SAORI o[W
#define SAORI_VERSIONSTRING_10 _T("SAORI/1.0")
#define SAORI_VERSIONSTRING SAORI_VERSIONSTRING_10

// SAORI wp[}N
#define DEFINE_SAORI_FUNCTION(name) SAORIResult saori_function_##name(const SAORIInput &in, SAORIOutput &out)
#define REGISTER_SAORI_FUNCTION(name) functions[_T(#name)] = saori_function_##name

#define VERIFY_ARGUMENT(count) if (in.args.size() < count) { return SAORIRESULT_BAD_REQUEST; }
#define CHECK_ARGUMENT(count) in.args.size() == count

// GNX|[g֐`
SAORIAPI int SAORICALL load(void *h, long len);
SAORIAPI int SAORICALL unload();
SAORIAPI void * SAORICALL request(void *h, long *len);

typedef std::basic_string<char_t> string_t;
typedef std::basic_istringstream<char_t> istringstream_t;
typedef std::basic_ostringstream<char_t> ostringstream_t;

typedef enum
{
	SAORIRESULT_OK = 200,
	SAORIRESULT_NO_CONTENT = 204,
	SAORIRESULT_BAD_REQUEST = 400,
	SAORIRESULT_INTERNAL_SERVER_ERROR = 500,
} SAORIResult;

typedef enum
{
	SAORICHARSET_SHIFT_JIS = 932,
	SAORICHARSET_UTF_8 = 65001,
	SAORICHARSET_EUC_JP = 20932,
	SAORICHARSET_ISO_2022_JP = 50220,
} SAORICharset;

class SAORIInput;
class SAORIOutput;

// SAORI ֐
typedef boost::function<SAORIResult (const SAORIInput &, SAORIOutput &)> SAORIFunction;

// SAORI NGXg
class SAORIInput
{
public:
    SAORIInput(string_t req) { deserialize(req); }
	bool deserialize(string_t req);
public:
	SAORICharset charset;
	string_t command;
    string_t function;
	std::vector<string_t> args;
	std::map<string_t, string_t> opts;
};

// SAORI X|X
class SAORIOutput
{
public:
	string_t serialize();
public:
	SAORICharset charset;
	SAORIResult result_code;
	string_t result;
	std::vector<string_t> values;
	std::map<string_t, string_t> opts;
};

// SAORI C
class SAORI
{
public:
	SAORI(const std::string &path)
	{
#ifdef _SAORI_UNICODE
		setlocale(LC_ALL, "Japanese");
		saori_path = MultibyteToUnicode(SAORICHARSET_SHIFT_JIS, path);
#else
		saori_path = path;
#endif /* _SAORI_UNICODE */
	}
	// ׂ֐
	bool load();
	bool unload();
    // NGXg֐
	std::string request(std::string req);
private:
	string_t saori_path;
    std::map<string_t, SAORIFunction> functions;
public:
	static string_t ResultToString(SAORIResult result);
	static string_t CharsetToString(SAORICharset charset);
	static SAORICharset StringToCharset(const std::string &str);
#ifdef _SAORI_UNICODE
	static std::wstring MultibyteToUnicode(SAORICharset charset, const std::string &str);
	static std::string UnicodeToMultibyte(SAORICharset charset, const std::wstring &wstr);
#endif /* _SAORI_UNICODE */
};

// ^ϊ + R[hϊwp[
namespace conv_helper
{
    template<class Target, class Source, bool IsConvertible>
    struct conv_op
    {
        Target operator()(const Source &src) const
        {
            return boost::lexical_cast<Target>(src);
        }
    };

    template<class Target, class Source>
    struct conv_op<Target, Source, true>
    {
        Target operator()(const Source &src) const
        {
            return static_cast<Target>(src);
        }
    };

    // R[h݂͓ꉻőΏ
    template<>
    struct conv_op<std::string, std::wstring, false>
    {
        std::string operator()(const std::wstring &src) const
        {
            return SAORI::UnicodeToMultibyte(SAORICHARSET_SHIFT_JIS, src);
        }
    };

    template<>
    struct conv_op<std::wstring, std::string, false>
    {
        std::wstring operator()(const std::string &src) const
        {
            return SAORI::MultibyteToUnicode(SAORICHARSET_SHIFT_JIS, src);
        }
    };
    
    // __int8ϊłȂ̂
    template<>
    struct conv_op<signed __int8, string_t , false>
    {
        signed __int8 operator()(const string_t &src) const
        {
            return static_cast<signed __int8>(boost::lexical_cast<signed __int16>(src));
        }
    };

    template<>
    struct conv_op<unsigned __int8, string_t , false>
    {
        unsigned __int8 operator()(const string_t &src) const
        {
            return static_cast<unsigned __int8>(boost::lexical_cast<unsigned __int16>(src));
        }
    };
}

template<class Target, class Source>
Target conv(const Source &src)
{
    return conv_helper::conv_op<Target, Source, boost::is_convertible<Source, Target>::value>()(src);
}

SAORIResult execute(const SAORIInput &, SAORIOutput &);